import Base.show

"""
    PartiallySeparableNLPModel{ T, S, G, M <: AbstractNLPModel{T, S}, Meta <: AbstractNLPModelMeta{T, S}} <: AbstractPartiallySeparableNLPModel{T, S}

A partitioned quasi-Newton `NLPModel`.
A `PartiallySeparableNLPModel` has field:

* `meta`: gather information about the `PartiallySeparableNLPModel`;
* `part_data`: allocate the partitioned structures required by a partitioned quasi-Newton trust-region method;
* `nlp`: the original `NLPModel`;
* `counters`: count how many standards methods of `NLPModels` are called.
"""
mutable struct PartiallySeparableNLPModel{
  T,
  S,
  G,
  M <: AbstractNLPModel{T, S},
  Meta <: AbstractNLPModelMeta{T, S},
} <: AbstractPartiallySeparableNLPModel{T, S}
  meta::Meta
  part_data::PartitionedDataTRPQN{G, T}
  nlp::M
  counters::NLPModels.Counters
end

function PartiallySeparableNLPModel(nlp::SupportedNLPModel; kwargs...)
  n = nlp.meta.nvar
  x0 = nlp.meta.x0
  ex = get_expression_tree(nlp)
  part_data_plbfgs = build_PartitionedDataTRPQN(ex, n; x0 = x0, kwargs...)
  meta = nlp.meta
  counters = NLPModels.Counters()
  PartiallySeparableNLPModel(meta, part_data_plbfgs, nlp, counters)
end

function update_nlp(
  nlp::PartiallySeparableNLPModel,
  x::AbstractVector{T},
  s::AbstractVector{T};
  kwargs...,
) where {T}
  update_nlp!(nlp, x, s; kwargs...)
  return Matrix(get_pB(nlp.part_data))
end

function Mod_PQN.update_nlp!(
  nlp::PartiallySeparableNLPModel,
  x::AbstractVector{T},
  s::AbstractVector{T};
  kwargs...,
) where {T}
  part_data = nlp.part_data
  update_PQN!(part_data, x, s; kwargs...)
  return part_data
end

Mod_ab_partitioned_data.get_pB(nlp::PartiallySeparableNLPModel) = get_pB(nlp.part_data)

Mod_ab_partitioned_data.get_py(nlp::PartiallySeparableNLPModel) = get_py(nlp.part_data)
